! This program is released under the MIT license. See "LICENSE.txt".
!
! Fast HyperVolume (FHV) by
! Takeshi WATANABE, Tomoaki TATSUKAWA, and Akira OYAMA.
!
! Please refer to the following paper when you publish results of
! investigations using the FHV.
!
! Takeshi Watanabe, Tomoaki Tatsukawa, and Akira Oyama, On the fast hypervolume
! calculation method, Proceedings of 2015 IEEE congress on evolutionary
! computation, pp. 965-969, 25th-28th May, 2015, Sendai, Japan.
!
! See "README.txt" for usage instructions.

!_____________________________________________________________________________!
!_____________________________________FHV_____________________________________!
program fhv
  !use omp_lib
  implicit none
  character(128) :: file
  character(4096) :: line
  integer :: dim,len,l,enddim,div,d,nf
  real(8),dimension(:),allocatable :: data
  real(8),dimension(:,:),allocatable :: datas
  real(8) :: hv,mhv
  !___for_hv_and_wfg___
  !integer :: i,j,unit
  !real(8) :: dresult
  !real(8),dimension(:),allocatable :: hvs
  !character(10) :: fname
  !character(1024) :: cmd
  
  if(iargc()/=1)call usage()
  
  call getarg(1,file)
  open(1,file=trim(file),status="old",form="formatted",err=100)
  read(1,'(a)')line
  close(1)
  dim=0
  do while(.true.)
     dim=dim+1
     allocate(data(dim))
     read(line,*,end=5,err=400)data
     deallocate(data)
  end do
5 dim=dim-1
  deallocate(data)
  allocate(data(dim))
  
  open(1,file=trim(file),status="old",form="formatted",err=100)
  len=0
  do while(.true.)
     read(1,*,err=10,end=10)data
     len=len+1
  end do
  
10 close(1)
  deallocate(data)
  
  allocate(datas(dim,len))
  open(1,file=trim(file),status="old",form="formatted",err=100)
  do l=1,len
     read(1,*)datas(:,l)
  end do
  close(1)
  
  d      = 1
  enddim = dim
  div    = 2
  nf     = 0
  hv     = 0.d0
  mhv    = 0.d0
  call divide(d,div,dim,enddim,len,datas,nf,hv,mhv)
  
  !___for_hv_and_wfg___
  ! allocate(hvs(div**(enddim-d+1)))
  ! !$OMP PARALLEL DO SCHEDULE(dynamic) PRIVATE(fname,cmd,unit,dresult)
  ! do i=1,div**(enddim-d+1)
  !    write(fname,'(1i6.6)')i
  !    fname(7:10)=".dat"
  !    !___hv___
  !    !cmd='./hv -r "'
  !    !do j=1,dim
  !    !   cmd(8+j*2:8+j*2+1)="1 "
  !    !end do
  !    !cmd(8+dim*2+1:8+dim*2+2)='" '
  !    !cmd(8+dim*2+2+1:8+dim*2+2+10)=fname
  !    !cmd(19+dim*2+2:19+dim*2+18)=" > "//fname//".res"
  !    !unit=omp_get_thread_num()+10
  !    !call system(cmd)
  !    !open(unit,file=fname//".res",status="old")
  !    !read(unit,*)dresult
  !    !close(unit)
  !    !hvs(i)=dresult
  !    !___wfg___
  !    !cmd='./wfg2 '
  !    !cmd(7+1:7+10)=fname
  !    !do j=1,dim
  !    !   cmd(16+j*2:16+j*2+1)=" 1"
  !    !end do
  !    !cmd(16+dim*2+2:16+dim*2+18)=" > "//fname//".res"
  !    !unit=omp_get_thread_num()+10
  !    !call system(cmd)
  !    !open(unit,file=fname//".res",status="old")
  !    !read(unit,*)dresult
  !    !close(unit)
  !    !hvs(i)=dresult
  !    !~~~
  ! end do
  ! !$OMP END PARALLEL DO
  ! write(*,'(1e38.30)')sum(hvs)
  !~~~~~~~~~~~~~~~~~~~~
  
  stop
  
100 write(*,*)"No such file: "//trim(file)//"."; stop
400 write(*,*)"Unrecognized data format."; stop
  
  stop
end program fhv
!~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~!

!___________________________________DIVIDE____________________________________!
recursive subroutine divide(d,div,dim,enddim,len,datas,nf,hv,mhv)
  implicit none
  integer,intent(in) :: d,div,dim,enddim,len
  integer,intent(inout) :: nf
  real(8),dimension(dim,len),intent(inout) :: datas
  real(8),intent(inout) :: hv,mhv
  real(8),dimension(div) :: data
  integer,dimension(0:div) :: dpos
  logical,dimension(len) :: nd
  integer :: i,itmp,j,k,l,nlen
  logical :: ldmd,lsame,lndmd
  type :: eachdim
     real(8),dimension(:,:),allocatable :: datas
     integer :: len
  end type eachdim
  type(eachdim),dimension(div) :: objd
  character(10) :: fname
  character(1024) :: line
  real(8) :: p
  integer :: m,tnlen
  
  call quicksort(dim,len,datas,d,1,len)
  dpos(0)=0
  nd=.true.
  do i=1,div
     if(i==div)then
        data(div)=1.d0
        dpos(div)=len
     end if
     nlen=0
     do j=1,dpos(i-1)
        if(nd(j))then
           nlen=nlen+1
        end if
     end do
     allocate(objd(i)%datas(dim,nlen+len))
     nlen=0
     do j=1,dpos(i-1)
        if(nd(j))then
           nlen=nlen+1
           do k=1,dim
              if(k==d)then
                 objd(i)%datas(k,nlen)=data(i-1)+(1.d0-data(i))
              else
                 objd(i)%datas(k,nlen)=datas(k,j)
              end if
           end do
        end if
     end do
     if(i/=div)then
        do j=1,len
           tnlen=j
           do k=1,j-1
              if(.not.nd(k))then
                 tnlen=tnlen-1
                 cycle
              end if
              ldmd =.true.
              lndmd=.true.
              lsame=.true.
              
              do l=1,dim
                 if(l/=d.and.datas(l,j)>datas(l,k))then
                    ldmd =.false.
                    lsame=.false.
                    if(.not.lndmd.and..not.lsame)exit
                 else if(l/=d.and.datas(l,j)<datas(l,k))then
                    lndmd=.false.
                    lsame=.false.
                    if(.not.ldmd .and..not.lsame)exit
                 end if
              end do
              
              if(ldmd .and..not.lsame)then
                 nd(k)=.false.
                 tnlen=tnlen-1
              else if(lndmd.and..not.lsame.or.lsame)then
                 nd(j)=.false.
                 tnlen=tnlen-1
              end if
           end do
           p=1.d0
           if(tnlen<=(j*(div-i+1)-len)*p)then
              itmp=j
              if(datas(d,itmp)==datas(d,itmp+1))then
                 cycle
              else
                 dpos(i)=itmp
                 data(i)=(datas(d,itmp)+datas(d,itmp+1))/2.d0
              end if
              exit
           end if
        end do
     end if
     objd(i)%len=nlen+dpos(i)-dpos(i-1)
     do j=dpos(i-1)+1,dpos(i)
        nlen=nlen+1
        do k=1,dim
           if(k==d)then
              objd(i)%datas(k,nlen)=datas(k,j)+(1.d0-data(i))
           else
              objd(i)%datas(k,nlen)=datas(k,j)
           end if
        end do
     end do
  end do
  
  if(d==enddim)then
     do i=1,div
        nf=nf+1
        write(fname,'(1i6.6)')nf
        fname(7:10)=".dat"
        open(1,file=fname,form="formatted")
        !___for_wfg___
        !write(1,'(a)')'#'
        !~~~~~~~~~~~~~
        do j=1,objd(i)%len
           line=""
           do k=1,dim
              write(line((k-1)*23+1:k*23),'(1e23.15)')objd(i)%datas(k,j)
           end do
           write(1,'(a)')trim(adjustl(line))
        end do
        !___for_wfg___
        !write(1,'(a)')'#'
        !~~~~~~~~~~~~~
        close(1)
     end do
  else
     do i=1,div
        call divide(d+1,div,dim,enddim,objd(i)%len,objd(i)%datas,nf,hv,mhv)
     end do
  end if
  
  return
end subroutine divide
!~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~!

!__________________________________QUICKSORT__________________________________!
recursive subroutine quicksort(dim,len,datas,d,left,right)
  implicit none
  integer,intent(in) :: dim,len
  real(8),dimension(dim,len),intent(inout) :: datas
  integer,intent(in) :: d,left,right
  integer :: i,j
  real(8) :: dpiv
  real(8),dimension(dim) :: tmp
  
  if(left<right)then
     i=left
     j=right
     dpiv=med(datas(d,i),datas(d,i+(j-i)/2),datas(d,j))
     do while(.true.)
        do while(datas(d,i)<dpiv)
           i=i+1
        end do
        do while(dpiv<datas(d,j))
           j=j-1
        end do
        if(i>=j)exit
        tmp=datas(:,i)
        datas(:,i)=datas(:,j)
        datas(:,j)=tmp
        i=i+1
        j=j-1
     end do
     call quicksort(dim,len,datas,d,left,  i-1)
     call quicksort(dim,len,datas,d, j+1,right)
  end if
  
  return
contains
  !____________________________________MED____________________________________!
  function med(x,y,z)
    implicit none
    real(8) :: med,x,y,z
    
    if(x<y)then
       if(y<z)then
          med=y
       else if(z<x)then
          med=x
       else
          med=z
       end if
    else
       if(z<y)then
          med=y
       else if(x<z)then
          med=x
       else
          med=z
       end if
    end if
    
    return
  end function med
  !~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~!
end subroutine quicksort
!~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~!

!____________________________________USAGE____________________________________!
subroutine usage()
  write(*,*)"Usage: ./fhv input_csv_file"
  stop
end subroutine usage
!~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~!
